/*----------------------------------------------------------------------
  File    : nass.c
  Contents: statistical tests for participation in neuron assemblies
  Author  : Christian Borgelt
  History : 2008.05.23 file created
            2008.05.27 test statistic for total data set added (sum)
            2008.06.12 statistics computation made a function
            2008.06.13 background rate estimation moved to this file
            2008.06.16 background rate estimation based on chi^2
            2008.08.27 conditional spike frequency analysis added
            2008.09.23 option -m added (max. number of other spikes)
            2008.09.24 frqs modified to count only excess spikes
            2008.09.25 background rate estimation simplified
            2008.09.27 output of shuffling statistics extended
            2008.10.01 bug in function cpc fixed (total p-value)
            2008.10.05 adapted to external array and math modules
            2008.10.06 conditional complexity frequencies test added
            2008.10.13 conditional pattern overlap test added
            2008.11.19 complete redesign with trial management module
            2008.11.24 main statistical test function added
            2009.01.31 splitting data into trials made optional
            2009.05.26 test statistic cxc added (excess complexity)
            2009.06.02 test statistic csw added (spike weights)
            2009.06.04 test statistics ccr, csr, and cwr added
            2011.11.30 mode without shuffling and option -z added
            2011.12.02 adapted to external utility module random
            2013.03.10 adapted to modified bsearch/bisect interface
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include "arrays.h"
#include "random.h"
#include "trials.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "nass"
#define DESCRIPTION "test whether neurons participate in an assembly"
#define VERSION     "version 3.10 (2014.10.24)        " \
                    "(c) 2008-2014   Christian Borgelt"

/* --- error codes --- */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* file open  failed */
#define E_FREAD      (-3)       /* file read  failed */
#define E_FWRITE     (-4)       /* file write failed (unused) */
#define E_OPTION     (-5)       /* unknown option */
#define E_OPTARG     (-6)       /* missing option argument */
#define E_ARGCNT     (-7)       /* wrong number of arguments */
#define E_PARAM      (-8)       /* invalid parameter value */
#define E_STAT       (-9)       /* invalid name of statistic */
#define E_IDENT     (-10)       /* invalid identifier (not integer) */
#define E_TRCNT     (-11)       /* no trial found (empty file) */
#define E_TRLEN     (-12)       /* trial lengths differ */
#define E_UNKNOWN   (-13)       /* unknown error */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_OPTION   -5 */  "unknown option -%c",
  /* E_OPTARG   -6 */  "missing option argument",
  /* E_ARGCNT   -7 */  "wrong number of arguments",
  /* E_PARAM    -8 */  "invalid parameter value %g",
  /* E_STAT     -9 */  "invalid name of statistic %s",
  /* E_IDENT   -10 */  "invalid identifier (not integer)",
  /* E_TRCNT   -11 */  "no trial found (empty file)",
  /* E_TRLEN   -12 */  "trial lengths differ: %d %d",
  /* E_UNKNOWN -13 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static const char *prgname;     /* program name for error messages */

static TRIALSET *trset = NULL;  /* trial set */
static FILE     *file  = NULL;  /* input file */

/*----------------------------------------------------------------------
  Statistics Evaluation
----------------------------------------------------------------------*/

static void header (const char *fmt)
{                               /* --- print ext. statistics header */
  printf(fmt, "minimum");       /* minimum */
  printf(fmt,  "1%");           /* 0.01 quantile */
  printf(fmt,  "5%");           /* 0.05 quantile */
  printf(fmt, "25%");           /* 0.25 quantile */
  printf(fmt, "50%");           /* 0.50 quantile */
  printf(fmt, "75%");           /* 0.75 quantile */
  printf(fmt, "95%");           /* 0.95 quantile */
  printf(fmt, "99%");           /* 0.99 quantile */
  printf(fmt, "maximum");       /* maximum */
  printf(fmt, "actual");        /* actual value */
  printf(" spikes");            /* spike count */
}  /* header() */

/*--------------------------------------------------------------------*/

static void report (const char *fmt,
                    double *stats, int n, double act, int spikes)
{                               /* --- report statistics information */
  printf(fmt, stats[0]);              /* minimum */
  printf(fmt, stats[(int)(0.01*n)]);  /* 0.01 quantile */
  printf(fmt, stats[(int)(0.05*n)]);  /* 0.05 quantile */
  printf(fmt, stats[(int)(0.25*n)]);  /* 0.25 quantile */
  printf(fmt, stats[(int)(0.50*n)]);  /* 0.50 quantile */
  printf(fmt, stats[(int)(0.75*n)]);  /* 0.75 quantile */
  printf(fmt, stats[(int)(0.95*n)]);  /* 0.95 quantile */
  printf(fmt, stats[(int)(0.99*n)]);  /* 0.99 quantile */
  printf(fmt, stats[n-1]);            /* maximum */
  printf(fmt, act);                   /* actual value */
  printf(" %6d", spikes);             /* number of spikes */
}  /* report() */

/*--------------------------------------------------------------------*/

static char sigind (double p)
{                               /* --- get significance indicator */
  if (p < 1e-6) return '#';     /* '#' means significant at 1e-6 */
  if (p < 0.01) return '*';     /* '*' means significant at 1% */
  return (p < 0.05) ? '<' : ' ';/* '<' means significant at 5% */
}  /* sigind() */

/*--------------------------------------------------------------------*/

static int test (TRIALSET *trset, int nid, int runs, double corr,
                 int code, int param, int npos, int ext)
{                               /* --- execute statistical test */
  int      i, k, n;             /* loop variables, buffer */
  int      s, spikes;           /* (total) number of spikes */
  double   *stats;              /* train statistic values */
  double   *sums;               /* sums of statistic values */
  TRSTATFN *statfn;             /* test statistic to compute */
  double   x, y, total;         /* (total value of) test statistic */
  double   p;                   /* computed p-value */

  /* --- initialize --- */
  stats = (double*)malloc((size_t)(runs +runs) *sizeof(double));
  if (!stats) return -1;        /* allocate the statistics arrays */
  sums  = stats +runs;          /* and organize them */
  total = spikes = 0;           /* initialize the statistic total */
  dbl_clear(sums,(size_t)runs); /* and clear the statistic sums */
  if (runs <= 0) {              /* if no shuffling */
    printf("neuron   value");   /* print a header for the output */
    if (ext) printf(" spikes");}/* add spike count as extension */
  else {                        /* if to copare to shuffled data */
    printf("neuron   p-value ");/* print a header for the output */
    if (ext) header(" %11s");   /* add extended output header */
  }                             /* (quantiles etc.) */
  printf("\n");                 /* terminate the header line */

  /* --- execute tests for neurons --- */
  statfn = trs_statfn(code);    /* get the statistic function */
  if (nid > 0) i = (n = nid)-1; /* get the neuron loop limits */
  else       { i = 0; n = trs_max(trset); }
  while (++i <= n) {            /* traverse the neurons to analyze */
    s = trs_prepare(trset, i);  /* prepare for the current neuron */
    if (s < 0) { free(stats); return -1; }
    spikes += s;                /* sum the number of spikes */
    x = (s > 0) ? statfn(trset, param, npos) : 0;
    total += x;                 /* compute and sum the test statistic */
    if (runs <= 0) {            /* if no shuffling requested, */
      printf("%6d  %11.6f", i, x);   /* report the statistic value */
      if (ext) printf(" %6d", s); }  /* and the number of spikes */
    else {                      /* if to compare to shuffled data */
      for (p = 0, k = runs; --k >= 0; ) {
        trs_shuffle(trset, corr, drand);  /* execute shuffling runs */
        stats[k] = y = (s > 0) ? statfn(trset, param, npos) : 0;
        sums[k] += y;           /* compute and sum test statistic */
        if (y >= x) p++;        /* count the values exceeding or */
      }                         /* meeting the actual statistic value */
      p /= runs;                /* compute and print the p-value */
      printf("%6d  %8.6f%c", i, p, sigind(p));
      if (ext) {                /* if extended output requested */
        dbl_heapsort(stats, (size_t)runs, +1);
        report(" %11.6f", stats, runs, x, s);
      }                         /* sort statistic values and */
    }                           /* report quantiles of statistic */
    printf("\n");               /* and the actual statistic value, */
  }                             /* then terminate the output line */

  /* --- total statistics --- */
  if ((runs > 0) && (nid <= 0)){/* if all neurons were checked */
    dbl_heapsort(sums, (size_t)runs, +1); /* sort summed statistics */
    for (p = 0.0, k = 0; k < runs; k++)
      if (sums[k] >= total) p++;/* compute the p-value as the number */
    p /= (double)runs;          /* of runs at least the total value */
    printf("total   %8.6f%c", p, sigind(p));
    if (ext)                    /* if extended output requested */
      report(" %11.4f", sums, runs, total, spikes);
    printf("\n");               /* report quantiles of statistic */
  }                             /* and terminate the output line */

  /* --- clean up --- */
  free(stats);                  /* free the allocated memory */
  return 0;                     /* return whether an error occurred */
}  /* test() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print an error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */

  assert(prgname);              /* check the program name */
  if (code < E_UNKNOWN) code = E_UNKNOWN;
  if (code < 0) {               /* if to report an error, */
    msg = errmsgs[-code];       /* get the error message */
    if (!msg) msg = errmsgs[-E_UNKNOWN];
    fprintf(stderr, "\n%s: ", prgname);
    va_start(args, code);       /* get variable arguments */
    vfprintf(stderr, msg, args);/* print error message */
    va_end(args);               /* end argument evaluation */
    fputc('\n', stderr);        /* terminate the output line */
  }
  #ifndef NDEBUG                /* if debug version */
  if (trset) trs_delete(trset); /* free memory and close files */
  if (file && (file != stdin)) fclose(file);
  #endif
  exit(code);                   /* abort the program */
}  /* error() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, n, k = 0;          /* loop variables */
  char    *s;                   /* to traverse the options */
  char    **optarg = NULL;      /* option argument */
  char    *fname   = NULL;      /* name of input gdf file */
  char    *sname   = "cpc";     /* name of statistic to compute */
  double  offset   = 0;         /* offset for time binning */
  double  width    = 1;         /* width  of  time bins */
  int     split    = 1;         /* flag for splitting into trials */
  int     nid      = 0;         /* single neuron id */
  int     stat     = 0;         /* identifier of statistic */
  int     min      = 1;         /* min. number of spikes of neuron */
  int     max      = 0;         /* max. spikes of other neurons */
  int     runs     = 1000;      /* number of shuffling runs */
  double  corr     = -1.0;      /* correction for complexity */
  int     expo     = 1;         /* exponent for freqs./complexity */
  int     npos     = 0;         /* flag for normalization mode */
  int     ext      = 0;         /* flag for extended output */
  long    seed;                 /* seed value for random numbers */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  if (argc <= 1) {              /* if no arguments are given */
    printf("usage: %s [options] file\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-o#    offset for time binning                 "
                  "(default: %g)\n", offset);
    printf("-w#    width  of  time bins                    "
                  "(default: %g)\n", width);
    printf("-t     do not split into trials\n");
    printf("-i#    index of neuron to check                "
                  "(default: check all)\n");
    printf("-s#    statistic to compute                    "
                  "(default: cpc)\n");
    printf("       bre: background rate estimation\n");
    printf("       cpc: conditional pattern complexity\n");
    printf("       cxc: conditional excess complexity\n");
    printf("       ccf: conditional complexity frequency\n");
    printf("       ccr: conditional complexity frequency ratio\n");
    printf("       cff: conditional Fano factor\n");
    printf("       csf: conditional spike frequency\n");
    printf("       csr: conditional spike frequency ratio\n");
    printf("       csw: conditional spike weight\n");
    printf("       cwr: conditional spike weight ratio\n");
    printf("       cpo: conditional pattern overlap\n");
    printf("-x#    power of complexities, frequencies etc. "
                  "(default: %d, all but bre)\n", expo);
    printf("-y#    max. number of spikes of other neurons  "
                  "(default: %d, only bre)\n", max);
    printf("-m#    min. number of spikes for computation   "
                  "(default: %d)\n", min);
    printf("-n#    number of spike shuffling runs          "
                  "(default: %d)\n", runs);
    printf("-S#    seed value for random numbers           "
                  "(default: time)\n");
    printf("-c#    correction for pattern complexity       "
                  "(default: uniform)\n");
    printf("-q     extended output (print quantiles)       "
                  "(default: only p-values)\n");
    printf("-z     normalize with number of positive terms "
                  "(default: all terms)\n");
    printf("file   input file in gdf format "
                  "(two columns: neuron id and spike time)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  seed = (long)time(NULL);      /* get the default seed value */
  for (i = k = 0; ++i < argc; ){/* traverse the arguments */
    s = argv[i];                /* get the next argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        switch (*s++) {         /* evaluate switches */
          case 'o': offset =      strtod(s, &s);    break;
          case 'w': width  =      strtod(s, &s);    break;
          case 't': split  = 0;                     break;
          case 'i': nid    = (int)strtol(s, &s, 0); break;
          case 's': optarg = &sname;                break;
          case 'm': min    = (int)strtol(s, &s, 0); break;
          case 'n': runs   = (int)strtol(s, &s, 0); break;
          case 'S': seed   =      strtol(s, &s, 0); break;
          case 'c': corr   =      strtod(s, &s);    break;
          case 'q': ext    = -1;                    break;
          case 'y': max    = (int)strtol(s, &s, 0); break;
          case 'x': expo   = (int)strtol(s, &s, 0); break;
          case 'z': npos   = -1;                    break;
          default : error(E_OPTION, *--s);          break;
        }                       /* set option variables */
        if (optarg && *s) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-options */
        case  0: fname = s;       break;
        default: error(E_ARGCNT); break;
      }                         /* note filename */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (k != 1) error(E_ARGCNT);  /* and the number of arguments */
  if (offset <  0) error(E_PARAM, offset);
  if (width  <= 0) error(E_PARAM, width);
  if (runs   <  0) error(E_PARAM, (double)runs);
  if (nid    <  0) error(E_PARAM, (double)nid);
  if      (strcmp(sname, "bre") == 0) stat = TRS_BRE;
  else if (strcmp(sname, "cpc") == 0) stat = TRS_CPC;
  else if (strcmp(sname, "cxc") == 0) stat = TRS_CXC;
  else if (strcmp(sname, "ccf") == 0) stat = TRS_CCF;
  else if (strcmp(sname, "ccr") == 0) stat = TRS_CCR;
  else if (strcmp(sname, "cff") == 0) stat = TRS_CFF;
  else if (strcmp(sname, "csf") == 0) stat = TRS_CSF;
  else if (strcmp(sname, "csr") == 0) stat = TRS_CSR;
  else if (strcmp(sname, "csw") == 0) stat = TRS_CSW;
  else if (strcmp(sname, "cwr") == 0) stat = TRS_CWR;
  else if (strcmp(sname, "cpo") == 0) stat = TRS_CPO;
  else error(E_STAT, sname);    /* decode the statistic name */

  /* --- read input file --- */
  t = clock();                  /* start timer for the reading */
  if ((strcmp(fname,  "") == 0) /* if no file name given, use stdin */
  ||  (strcmp(fname, "-") == 0)) {
    file = stdin; fname = "<stdin>"; }
  else                          /* if a file name is given, */
    file = fopen(fname, "r");   /* open the file for reading */
  if (!file) error(E_FOPEN, fname);
  fprintf(stderr, "reading %s ... ", fname);
  trset = trs_create(0);        /* create an empty trial set */
  if (!trset) error(E_NOMEM);   /* and read it from the file */
  k = trs_read(trset, file, split, offset, width);
  if (k <  -3) error(E_UNKNOWN);/* check for an error and */
  if (k <= -3) error(E_IDENT);  /* evaluate the error code */
  if (k <= -2) error(E_FREAD, fname);
  if (k <   0) error(E_NOMEM);
  if (file != stdin) fclose(file);
  file = NULL;                  /* close the input file */
  n = trs_cnt(trset);           /* get and the number of trials */
  if (n <= 0) error(E_TRCNT);   /* and check for at least one trial */
  k = tr_cnt(trs_get(trset,0)); /* get the first trial length */
  i = trs_diff(trset);          /* check for equal length trials */
  if (i >= 0) {                 /* if trials have different lengths */
    fprintf(stderr, "\n%s warning: ", prgname);
    fprintf(stderr, "trials lengths differ (%d/%d)!\n", k, i);
    trs_crop(trset, -1);        /* print a warning message and */
  }                             /* crop the trials to equal length */
  fprintf(stderr, "[%d neuron(s), ", trs_max(trset));
  fprintf(stderr, "%d trial(s), %d bin(s)]", n, k);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- compute statistics --- */
  t = clock();                  /* start timer for the computation */
  if (min > 1)                  /* if min. number of spikes required, */
    trs_filter(trset, min);     /* filter the trials */
  rseed((unsigned)seed);        /* seed the random number generator */
  if (stat == TRS_BRE) expo = max;
  if (test(trset, nid, runs, corr, stat, expo, npos, ext) != 0)
    error(E_NOMEM);             /* compute the statistical tests */
  fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  #ifndef NDEBUG                /* if debug version */
  trs_delete(trset);            /* delete the trial set */
  #endif
  return 0;                     /* terminate the program */
}  /* main() */
